/*************************************************************************
Title:		AVR-RDS
Author:		Uwe Nagel
Date:		07. Juni 2003
Compiler:	AVR-GCC 3.3
Hardware:	Any AVR with enough memory
			RDS-Decoder like TDA7330 or SAA6576??
			HD44780 compatible LCD text display
Copyright:
	Dieses Programm ist geistiges Eigentum des Autors.
	Nur die private, nicht kommerzielle Nutzung ist gestattet.
	Jegliche kommerzielle Nutzung bedarf der ausdrcklichen 
	Zustimmung des Autors.
**************************************************************************/


#include <stdlib.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <string.h>
#include "lcd.h"
#include "syndrom.h"
/*
** constant definitions
*/

#define adctest 0
#define grouptest 0
#define showgroup -1

#define FLYSTART 2400
#define SCROLLSPEED 300
#define UART_BAUD_RATE 1200
#define XTAL_CPU 4332000
#define MINTAKT 60*XTAL_CPU/1024/256
//#define MINTAKT 200

#define BLOCKA 0
#define BLOCKB 1
#define BLOCKC 2
#define BLOCKD 3

#define GROUP0A 0
#define GROUP2A 4
#define GROUP4A 8
#define GROUP6A 12
#define GROUP8A 16


/*
** function prototypes
*/ 
void show_group_0(void);
void show_group_2(void);
void lcd_dec(uint8_t byte);
void lcd_hex(uint8_t byte);
void init_strings(void);

typedef struct
	{	int8_t h;
		uint8_t m;
		uint16_t s;
		uint8_t day;
		uint8_t month;
		uint8_t year;
	} zeit_t;


/*
** global variables
*/ 
unsigned long rcvd;
uint8_t synchronized;
volatile unsigned char groupready;
uint8_t blockcount;
uint8_t bitcount;
volatile int16_t flywheel=100;
uint16_t block[4];
const uint16_t offset[]={0x3f00,0x6600,0x5a00,0x6d00};
uint16_t RTstart=0;
volatile uint16_t RTscroll=100;
uint16_t PIcode;
uint16_t lastPI;
uint16_t PIshown;
char TPTA=' ';
char PS[8];
char RT[128];
zeit_t CT;
uint8_t PTY;
uint8_t line3;
uint8_t Portd_merker;

INTERRUPT(SIG_INTERRUPT1)
{
uint16_t syn;
uint16_t data;
uint16_t mask;
uint8_t flag;
PORTD|=0x80;
	/* read RDS-data*/
	rcvd=rcvd << 1;
	if( bit_is_set(PIND,PD2) ) rcvd|=0x40;
	
	/* if not synced, search for characteristical syndrom for Block A */
	if( !synchronized ){
		if( syndrom(rcvd)==0x3d8 ){
			synchronized=1;
			blockcount=0;
			bitcount=1;
		}
	}
	
	if( synchronized ){
		if( --bitcount==0 ){
			
			/* subtract offset and calculate syndrom */
			if( (blockcount==BLOCKC) && (block[BLOCKB] & 0x0800 )){	/* Grouptype B ? */
				rcvd^=0x0000d400;
			}
			else{
				rcvd^=offset[blockcount];
			}
			syn=syndrom(rcvd);
			data=rcvd>>16;

			/* error corection */
			flag=1;
			for( mask=0x8000; mask; mask>>=1 ){
				if( (syn & 0x21f) == 0x200 ){
					flag = 0;	data ^= mask; }
				syn <<= 1;
				if( flag && (syn & 0x400) )	syn ^= 0x1B9;
			}

			/* unrecoverable error ? */
			if( syn & 0x1f ){
				synchronized=0;	/* throw away complete group */
			}
			else{
				block[blockcount++]=data;
				bitcount=26;
				if(blockcount==4){
					blockcount=0;	groupready=1;
				}
				flywheel=FLYSTART;
			}
		}
	}
	if( flywheel>0 ) flywheel--;
	if( RTscroll ) RTscroll--;
PORTD&=0x7f;
}

uint8_t incdec( uint8_t x ){
	if((++x & 0x0f)==0x0a) x+=6;
	return x;
}

INTERRUPT(SIG_OVERFLOW0)
{ TCNT0=6;
//PORTD|=0x80;
	if( ++CT.s==16245 ){
		CT.s=0;
		CT.m=incdec(CT.m);
		if(CT.m==0x60){
			CT.m=0;
			CT.h=incdec(CT.h);
			if( CT.h==0x24 ) CT.h=0;
		}
	}
	//if( bit_is_clear(PIND,PD6) ) line3++;
//PORTD&=0x7f;
}

void show_group_0(void){
uint8_t i;
	lcd_gotoxy(0,0);
	for(i=0; i<8; i++) lcd_putc(PS[i]);
	lcd_gotoxy(0,2); lcd_puts("PI="); lcd_hex(PIcode>>8); lcd_hex(PIcode&0xff);
	lcd_puts(" PTY="); lcd_hex(PTY);
}

void show_group_2(void){
uint8_t i;
	lcd_gotoxy(0,1);
	for(i=0; i<16; i++) lcd_putc(RT[(RTstart+i)&0x7f]);
	lcd_gotoxy(15,0); lcd_putc(TPTA);
}

void init_strings(void)
{
	memset(RT,' ',128);
	strncpy(PS,"-no RDS-",8);
}

void lcd_nibble(uint8_t nn)
{
	if(nn>9) nn+=7; lcd_putc(0x30+nn);
}

void lcd_hex(uint8_t byte)
{
	lcd_nibble(byte>>4);  lcd_nibble(byte&0x0f);
}

void lcd_dec(uint8_t byte)
{
	lcd_nibble(byte/10);  lcd_nibble(byte%10);
}

static const char message1[] PROGMEM ="--- AVR-RDS ---\n";
static const char message2[] PROGMEM =" Version 0.2.03\n";
static const char message3[] PROGMEM ="  07. Mai 2004 \n";
static const char message4[] PROGMEM ="(c) by Uwe Nagel";


int main(void)
{
int Addr;
int Grouptype;
double f;
uint8_t PIchange;
char sss[16];
    PORTD=0x5d;		/* pull-ups enabled, except for outputs */
	DDRD=0xa2;      /* 1010 0010 PD7=Test PD5=LCD-E PD1=TxD */
	
	PORTB=0x0f;     /* LCD-Data with Pullups */
	DDRB=0xf0;      /* LCD-Data input */
	
	PORTC=0xff;
	DDRC=0xff;
	
	sbi(GIMSK,INT1);/* enable INT1 Interrupt */
	sbi(MCUCR,ISC11); sbi(MCUCR,ISC10);/* Interrupt on rising edge */
	
	TCCR0=3; /* Timer0 Prescaler=64 */
	sbi(TIMSK,TOIE0); /* enable Timer0 interrupt */

	synchronized=0;  bitcount=0;  blockcount=0;  groupready=0; RTscroll=3600; PIchange=0;
	init_strings();

    /* initialize display, cursor off */
    lcd_init(LCD_DISP_ON);

	/* Einschaltmeldung */
    lcd_clrscr();
    lcd_puts_p(message1);
    lcd_puts_p(message2);
    lcd_puts_p(message3);
    lcd_puts_p(message4);
	
	sei();/*global Interrupt enable */

	/* wait 3s */
	while(RTscroll);
    lcd_clrscr();

	for (;;) {                           /* loop forever */
		if( flywheel==0 ){
			init_strings();
			lcd_clrscr();
			flywheel=-1;
		}
		if( groupready ){
			groupready=0;

			PIcode=block[BLOCKA];
			if(PIcode!=lastPI){
				PIchange=1;
				lastPI=PIcode;
				continue;
			}
			if( PIchange ){
				PIshown=PIcode;
				init_strings();
				lcd_clrscr();
			}
			PIchange=0;
				
			Grouptype=block[BLOCKB]>>11;
#if grouptest
			/* show available groups */
			lcd_gotoxy(Grouptype>>1,3); lcd_putc('A'+(Grouptype & 1));
#endif
			switch( Grouptype ){
				case GROUP0A:	/* Group 0A: PS TA TP PTY */
					Addr=(block[BLOCKB] & 0x03)*2;
					PS[Addr]=block[BLOCKD]>>8;  PS[Addr+1]=block[BLOCKD];
					if( block[BLOCKB] & 0x0400 ){
						if( block[BLOCKB] & 0x0010 ) TPTA='*'; else TPTA='-';
					}
					else TPTA=' ';
					if( Addr==6 ) show_group_0();
					PTY=(block[BLOCKB]>>5)&0x1f;
					break;
				case GROUP2A:	/* Group 2A: RT */
					Addr=(block[BLOCKB] & 0x1f)*4;
					RT[Addr]=block[BLOCKC]>>8; RT[Addr+1]=block[BLOCKC];
					RT[Addr+2]=block[BLOCKD]>>8; RT[Addr+3]=block[BLOCKD];
				break;
				case GROUP4A:	/* Group 4A: CT */
					CT.h=((block[BLOCKC]<<4) + (block[BLOCKD]>>12)) & 0x1f;
					Addr=(block[BLOCKD]&0x1f)>>1;
					if( block[BLOCKD]&0x20 ) CT.h-=Addr; else CT.h+=Addr;
					if( CT.h<0 ) CT.h+=24;
					if( CT.h>23 ) CT.h-=24;
					CT.m=(block[BLOCKD]>>6)&0x3f;

					f=(float)((((int32_t)(block[BLOCKB]&0x03)<<16)|block[BLOCKC])>>1);
					CT.year=(uint8_t)((f-15078.2)/365.25);
					CT.month=(int)((f-14956.1-(int32_t)(CT.year*365.25))/30.6001);
					CT.day=(uint16_t)f-14956-(int32_t)(CT.year*365.25)-(int32_t)(CT.month*30.6001);
					if(CT.month>13){
						CT.month-=13;
						CT.year++;
					}
					else CT.month--;
					CT.year%=100;
				break;
#if showgroup>0
				case 2*showgroup:
					lcd_gotoxy(0,3);
					lcd_puts(itoa(showgroup,sss,10)); lcd_puts(": ");
					lcd_hex(block[BLOCKB]>>8); lcd_hex(block[BLOCKB]); 
					lcd_hex(block[BLOCKC]>>8); lcd_hex(block[BLOCKC]); 
					lcd_hex(block[BLOCKD]>>8); lcd_hex(block[BLOCKD]);
				break;
#endif
			}
		}
		if( RTscroll==0 ){
			RTscroll=SCROLLSPEED;
			show_group_2();
			RTstart++;
			lcd_gotoxy(9,0); lcd_dec(CT.h);
			lcd_putc(':'); lcd_dec(CT.m);
			lcd_gotoxy(0,3); lcd_dec(CT.day); lcd_putc('.'); 
			lcd_dec(CT.month); lcd_putc('.'); lcd_dec(CT.year);
		}
    }
}
